/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *									     *
 * Library:	libsubdiv2.5D						     *
 *									     *
 * Module:	subdiv2.5D_math.c					     *
 *									     *
 * Description:	mathematical functions					     *
 *									     *
 * Date:	20000329						     *
 *									     *
 * Author:	Francisco MORN BURGOS   (http://www.gti.ssr.upm.es./~fmb/)  *
 *		Grupo Tratamiento Imgenes, Univ. Politcnica Madrid, Spain  *
 *									     *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#include <math.h>

#include "subdiv2.5D_priv.h"



sdVectorPtr sdmMidpoint(sdVectorPtr pa, sdVectorPtr pb)
{
  static sdVector m;

  if (!pa || !pb)
    return NULL;

  m = *pa;
  INC_VEC(m, (*pb));
  MUL_VEC(m, 0.5F);
  return &m;
} /* sdmMidpoint */


static sdVectorPtr sdmBarycentre(sdTrianglePtr t)
{
  static sdVector b;

  if (!t)
    return NULL;

  b = t->vtx[0]->pos;
  INC_VEC(b, t->vtx[1]->pos);
  INC_VEC(b, t->vtx[2]->pos);
  DIV_VEC(b, 3.0F);
  return &b;
} /* sdmBarycentre */


sdVectorPtr sdmNormalise(sdVectorPtr pt)
{
  static sdVector ptn;
  float mod;

  if (!pt)
    return NULL;

  ptn = *pt;
  mod = ptn.x*ptn.x + ptn.y*ptn.y + ptn.z*ptn.z;
  if (mod != 0.0F && mod != 1.0F) {
    mod = (float) (1.0 / sqrt(mod));
    MUL_VEC(ptn, mod);
  }
  return &ptn;
} /* sdmNormalise */


float sdmDotProd(sdVectorPtr a, sdVectorPtr b)
{
  if (!a || !b)
    return 0.0F;

  return a->x*b->x + a->y*b->y + a->z*b->z;
} /* sdmDotProd */


sdVectorPtr sdmVecDiff(sdVectorPtr a, sdVectorPtr b)
{
  static sdVector v;

  if (!a || !b)
    return NULL;

  v = *a;
  DEC_VEC(v, (*b));
  return &v;
} /* sdmVecDiff */


sdVectorPtr sdmVecProd(sdVectorPtr a, sdVectorPtr b)
{
  static sdVector v;

  if (!a || !b)
    return NULL;

  v.x = a->y * b->z - a->z * b->y;
  v.y = a->z * b->x - a->x * b->z;
  v.z = a->x * b->y - a->y * b->x;
  return &v;
} /* sdmVecProd */   


float sdmVecNorm(sdVectorPtr pt)
{
  float mod;

  if (!pt)
    return -1.0F;

  mod = pt->x*pt->x + pt->y*pt->y + pt->z*pt->z;
  if (mod != 0.0F && mod != 1.0F)
    mod = (float) sqrt(mod);
  return mod;
} /* sdmVecNorm */


void sdmFillTriangleNormalAndK(sdTrianglePtr t, sdBoolean KisArea)
{
  int i;
  sdVector v[3];
  float mod;

  if (!t)
    return;

  for (i=0; i<3; i++)
    v[i] = t->vtx[i]->pos;
  DEC_VEC(v[1], v[0]);
  DEC_VEC(v[2], v[0]);
  t->normal = *sdmVecProd(v+1, v+2);
  mod = sdmVecNorm(&t->normal);
  if (mod != 0.0 && mod != 1.0)
    DIV_VEC(t->normal, mod);
  if (KisArea)
    t->K = mod / 2.0F;
  else
    t->K = sdmDotProd(&t->normal, v);
} /* sdmFillTriangleNormalAndK */


void sdmFillVertexNormal(sdPointPtr vtx, int n, sdPointPtr *nbr)
/* gvs was here */
{
  float arg, argstep, real, imag, **L1;
  sdVector t1, t2;
  int j;

  if (!vtx || !nbr)
    return;

  if (!(L1 = NEW(n, float *)))
    sdiError("sdmFillNormal: not enough memory for a set of (co)sines", -13);

  argstep = 2.0F * M_PI / n; 
  for (j=0, arg=0; j<n; j++, arg+=argstep) {
    if (!(L1[j] = NEW(2, float)))
      sdiError("sdmFillNormal: not enough memory for a pair of (co)sines", -13);
    L1[j][0] = (float) cos(arg);
    L1[j][1] = (float) sin(arg);
  }

  /* Coordinate t1x */
  for (j=0, real=imag=0; j<n; j++) {
    real += L1[j][0] * nbr[j]->pos.x;
    imag += L1[j][1] * nbr[j]->pos.x;
  }
  t1.x = 2*real;
  t2.x = 2*imag;
  
  /* Coordinate t1y */
  for (j=0, real=imag=0; j<n; j++) {
    real += L1[j][0] * nbr[j]->pos.y;
    imag += L1[j][1] * nbr[j]->pos.y;
  }
  t1.y = 2*real;
  t2.y = 2*imag;

  /* Coordinate t1z */
  for (j=0, real=imag=0; j<n; j++) {
    real += L1[j][0] * nbr[j]->pos.z;
    imag += L1[j][1] * nbr[j]->pos.z;
  }
  t1.z = 2*real;
  t2.z = 2*imag;

  vtx->normal = *sdmNormalise(sdmVecProd(&t1, &t2));
} /* sdmFillVertexNormal */


sdBoolean sdmCoplanarNeighbours(sdTrianglePtr t, float epsilon)
{
  int i;
  sdTrianglePtr ta;

  if (!t)
    return TRUE;

  for (i=0; i<3; i++)
    if ((ta = t->adt[i]) && sdmDotProd(&ta->normal, &t->normal) < epsilon)
      return FALSE;
  return TRUE;
} /* sdmCoplanarNeighbours */


sdVectorPtr sdmButterflyMidpoint(sdTrianglePtr t, int k, sdBoolean initCall)
{
  static sdVector blend;
  int n, i;
  static sdPointPtr nbr_t[12], nbr[8];
  sdTrianglePtr ta, taa;
  static int idx[3][8] = { { 0, 1, 2, 3, 4, 5, 6, 7},
			   { 1, 2, 0, 4, 5, 3, 8, 9},
			   { 2, 0, 1, 5, 3, 4,10,11} };

  /*
	          10     9
	          / \   / \ 
	         /   \ /   \
	        5-----2-----4	    5-----2-----4
	       / \   / \   / \	     \   / \   /
	      /   \ / t \ /   \	      \ /   \ /
	    11-----0-----1-----8       0--X--1
	          / \   / \	      / \   / \
	         /   \ /   \	     /   \ /   \
	        6-----3-----7	    6-----3-----7
  */

  if (!t)
    return NULL;

  if (initCall)
    for (n=0; n<3; n++) {
      nbr_t[n] = t->vtx[n];
      if (!(ta = t->adt[n])) {
	nbr_t[3+n] = NULL;
	ta = t;
	do {
	  i = IDX_VTX_IN_T(t->vtx[n], ta);
	  nbr_t[6+2*n] = ta->vtx[sdM1[i]];
	  ta = ta->adt[sdM1[i]];
	} while (ta);
	ta = t;
	do {
	  i = IDX_VTX_IN_T(t->vtx[sdP1[n]], ta);
	  nbr_t[7+2*n] = ta->vtx[sdP1[i]];
	  ta = ta->adt[i];
	} while (ta);
      } else {
	i = IDX_VTX_IN_T(t->vtx[n], ta);
	nbr_t[3+n] = ta->vtx[sdP1[i]];
	if (!(taa = ta->adt[i]))
	  nbr_t[6+2*n] = NULL;
	else
	  nbr_t[6+2*n] = taa->vtx[sdP1[IDX_VTX_IN_T(t->vtx[n], taa)]];
	if (!(taa = ta->adt[sdP1[i]]))
	  nbr_t[7+2*n] = NULL;
	else
	  nbr_t[7+2*n] = taa->vtx[sdP1[IDX_VTX_IN_T(nbr_t[3+n], taa)]];
      }
    }

  for (i=0; i<8; i++)
    nbr[i] = nbr_t[idx[k][i]];

  if (!nbr[3]) {
    if (!nbr[4] || !nbr[5] || !nbr[6] || !nbr[7]) {
      blend.x = (nbr[0]->pos.x+nbr[1]->pos.x) / 2.0F;
      blend.y = (nbr[0]->pos.y+nbr[1]->pos.y) / 2.0F;
      blend.z = (nbr[0]->pos.z+nbr[1]->pos.z) / 2.0F;
    } else {
      blend.x = ( 9 * (nbr[0]->pos.x + nbr[1]->pos.x)
		  - nbr[6]->pos.x - nbr[7]->pos.x ) / 16.0F;
      blend.y = ( 9 * (nbr[0]->pos.y + nbr[1]->pos.y)
		  - nbr[6]->pos.y - nbr[7]->pos.y ) / 16.0F;
      blend.z = ( 9 * (nbr[0]->pos.z + nbr[1]->pos.z)
		  - nbr[6]->pos.z - nbr[7]->pos.z ) / 16.0F;
    }
  } else if (!nbr[4] || !nbr[5] || !nbr[6] || !nbr[7]) {
    blend.x = ( 4 * (nbr[0]->pos.x + nbr[1]->pos.x)
		+ nbr[2]->pos.x + nbr[3]->pos.x ) / 10.0F;
    blend.y = ( 4 * (nbr[0]->pos.y + nbr[1]->pos.y)
		+ nbr[2]->pos.y + nbr[3]->pos.y ) / 10.0F;
    blend.z = ( 4 *(nbr[0]->pos.z + nbr[1]->pos.z)
		+ nbr[2]->pos.z + nbr[3]->pos.z ) / 10.0F;
  } else {
    blend.x = (8*(nbr[0]->pos.x+nbr[1]->pos.x) + 2*(nbr[2]->pos.x+nbr[3]->pos.x)
	       -nbr[4]->pos.x-nbr[5]->pos.x-nbr[6]->pos.x-nbr[7]->pos.x)/16.0F;
    blend.y = (8*(nbr[0]->pos.y+nbr[1]->pos.y) + 2*(nbr[2]->pos.y+nbr[3]->pos.y)
	       -nbr[4]->pos.y-nbr[5]->pos.y-nbr[6]->pos.y-nbr[7]->pos.y)/16.0F;
    blend.z = (8*(nbr[0]->pos.z+nbr[1]->pos.z) + 2*(nbr[2]->pos.z+nbr[3]->pos.z)
	       -nbr[4]->pos.z-nbr[5]->pos.z-nbr[6]->pos.z-nbr[7]->pos.z)/16.0F;
  }

  return &blend;
} /* sdmButterflyMidpoint */


sdVectorPtr sdmMatrixVecProd(sdVectorPtr m, sdVectorPtr v)
{
  static sdVector r;

  if (!m || !v)
    return NULL;

  r.x = m[0].x * v->x + m[1].x * v->y + m[2].x * v->z;
  r.y = m[0].y * v->x + m[1].y * v->y + m[2].y * v->z;
  r.z = m[0].z * v->x + m[1].z * v->y + m[2].z * v->z;
  return &r;
} /* sdmMatrixVecProd */


void sdmMatrixTrans(sdVectorPtr m, sdVectorPtr mt)
{
  if (!m || !mt)
    return;

  mt[0].x = m[0].x;  mt[0].y = m[1].x;  mt[0].z = m[2].x;
  mt[1].x = m[0].y;  mt[1].y = m[1].y;  mt[1].z = m[2].y;
  mt[2].x = m[0].z;  mt[2].y = m[1].z;  mt[2].z = m[2].z;
} /* sdmMatrixTrans */


static sdBoolean sdmPointInsideTriangle(sdVectorPtr pt, sdTrianglePtr t)
{
  double nx, ny, nz, u1, u2, v1, v2, u, v, a, b;

  if (!pt || !t)
    return FALSE;

  /* cf. "An efficient ray-polygon intersection" by Didier Badouel in	*/
  /* "Graphics Gems", pp. 390-393, whose second step ("intersecting the	*/
  /* polygon") is as follows:						*/
  /* 1.	project "t" onto one of the three planes (x,y), (y,z) or (z,x)	*/
  /*	depending on which "t" is "more parallel" to			*/
  nx = ABS(t->normal.x);  ny = ABS(t->normal.y);  nz = ABS(t->normal.z);
  if (nz >= nx && nz >= ny) {
    u1 = t->vtx[1]->pos.x - t->vtx[0]->pos.x;
    v1 = t->vtx[1]->pos.y - t->vtx[0]->pos.y;
    u2 = t->vtx[2]->pos.x - t->vtx[0]->pos.x;
    v2 = t->vtx[2]->pos.y - t->vtx[0]->pos.y;
    u  = pt->x		  - t->vtx[0]->pos.x;
    v  = pt->y		  - t->vtx[0]->pos.y;
  } else if (nx >= ny && nx >= nz) {
    u1 = t->vtx[1]->pos.y - t->vtx[0]->pos.y;
    v1 = t->vtx[1]->pos.z - t->vtx[0]->pos.z;
    u2 = t->vtx[2]->pos.y - t->vtx[0]->pos.y;
    v2 = t->vtx[2]->pos.z - t->vtx[0]->pos.z;
    u  = pt->y		  - t->vtx[0]->pos.y;
    v  = pt->z		  - t->vtx[0]->pos.z;
  } else if (ny >= nz && ny >= nx) {
    u1 = t->vtx[1]->pos.z - t->vtx[0]->pos.z;
    v1 = t->vtx[1]->pos.x - t->vtx[0]->pos.x;
    u2 = t->vtx[2]->pos.z - t->vtx[0]->pos.z;
    v2 = t->vtx[2]->pos.x - t->vtx[0]->pos.x;
    u  = pt->z		  - t->vtx[0]->pos.z;
    v  = pt->x		  - t->vtx[0]->pos.x;
  }
  /* 2.	calculate the normalised parametric coordinates of "pt" with	*/
  /*	respect to the projected triangle and verify that they both	*/
  /*	belong to the range [0,1] and that their sum does too		*/
  if (NEGLIGIBLE(u1)) {
    b = u / u2;
    if (b < 0 || b > 1)
      return FALSE;
    else
      a = (v - b*v2) / v1;
  } else {
    b = (u1*v - v1*u) / (u1*v2 - v1*u2);
    if (b < 0 || b > 1)
      return FALSE;
    else
      a = (u - b*u2) / u1;
  }
  return (a >= 0 && a+b <= 1);
} /* sdmPointInsideTriangle */


sdVectorPtr sdmCalculateDetail(sdMeshPtr m, sdPointPtr pt, sdVectorPtr dir)
{
  static sdVector tgt;
  sdVector org, ray, hit;
  int i;
  sdTrianglePtr t;
  double d, dmin;

  if (!m || !m->targetMesh || !pt)
    return NULL;

  if (m->useBarycentre) {
    org = m->barycentre;
    ray = *sdmVecDiff(&pt->pos, &org);
  } else {
    org = pt->pos;
    ray = *dir;
  }
  for (i=0, tgt=pt->pos, dmin=MOGOLLON;
       i<m->targetMesh->NmotherTriangles; i++) {
    t = m->targetMesh->motherTriangle[i];
    /* 0. determine whether "midpoint" lies inside "t"			*/
    d = t->K - sdmDotProd(&t->normal, &pt->pos);
    if (NEGLIGIBLE(d) && sdmPointInsideTriangle(&pt->pos, t))
      return &sdNullVector;
    /* cf. "An efficient ray-polygon intersection" by Didier Badouel in	*/
    /* "Graphics Gems", pp. 390-393:					*/
    /* 1. determine whether the ray intersects "t"'s plane		*/
    d = sdmDotProd(&t->normal, &ray);
    if (NEGLIGIBLE(d))	/* "ray" parallel to "t"'s plane */
      continue;
    d = (t->K - sdmDotProd(&t->normal, &org)) / d;
    if (ABS(d) >= ABS(dmin) ||		/* closer solution found before	*/
	m->useBarycentre && d <= 0)	/* intersection point behind "org" */
      continue;
    /* 2. determine whether the intersection point lies inside "t"	*/
    hit = ray;
    MUL_VEC(hit, (float) d);
    INC_VEC(hit, org);
    if (sdmPointInsideTriangle(&hit, t)) {
      tgt = hit;
      dmin = d;
    }
  }
  DEC_VEC(tgt, pt->pos);
  return &tgt;
} /* sdmCalculateDetail */


void sdmBuildFrameForTriangleEdge(sdTrianglePtr t, int i, sdVectorPtr v)
{
  sdTrianglePtr ta;
  sdVector van;

  if (!t || !v)
    return;

  /* "v[0]" will be the (area-weighted) average of the normals of "t"	*/
  /* and "t->adt[i]" (the two triangles sharing the edge), "v[1]" will	*/
  /* follow the edge and point from "t->vtx[i]" to "t->vtx[i+1]", and	*/
  /* "v[2] = v[0] x v[1]"; all vectors will be normalised		*/
  v[0] = t->normal;
  if (ta = t->adt[i]) {
    MUL_VEC(v[0], t->K);
    van = ta->normal;
    MUL_VEC(van, ta->K);
    INC_VEC(v[0], van);
    v[0] = *sdmNormalise(v);
  }
  v[1] = t->vtx[sdP1[i]]->pos;
  DEC_VEC(v[1], t->vtx[i]->pos);
  v[1] = *sdmNormalise(v+1);
  v[2] = *sdmVecProd(v, v+1);
} /* sdmBuildFrameForTriangleEdge */


void sdmEncodeDetail(sdMeshPtr m, int nbits, sdVectorPtr delta,
		     CODEWORD *codedelta)
{
  CODEWORD cd, maskN, maskT1, maskT2, mask;
  double vN, vT1, vT2, pw;
  int i;
  sdBoolean zN, zT1, zT2;

  *codedelta = 0;
  if (!m || !nbits)
    return;

  if (nbits > NBPDN) nbits = NBPDN;
  cd = 0; maskN = MASK_SN;
  if (delta->x > 0)	vN = +delta->x;
  else {cd |= maskN;	vN = -delta->x;}
  for (i=1, maskN>>=1, pw=0.5*m->maxDetailN, zN=TRUE;
       i<nbits; i++, maskN>>=1, pw*=0.5)
    if (vN > pw) {vN -= pw; cd |= maskN; zN = FALSE;}
  if (zN) cd &= ~MASK_SN;	/* avoid "-0" */
  nbits -= NBPDN-NBPDT;
  if (nbits > 1) {
    if (nbits > NBPDT) nbits = NBPDT;
    maskT1 = MASK_ST1;
    if (delta->y > 0)	vT1 = +delta->y;
    else {cd |= maskT1;	vT1 = -delta->y;}
    maskT2 = MASK_ST2;
    if (delta->z > 0)	vT2 = +delta->z;
    else {cd |= maskT2;	vT2 = -delta->z;}
    for (i=1, maskT1>>=1, maskT2>>=1, pw=0.5*m->maxDetailT, zT1=zT2=TRUE;
	 i<nbits; i++, maskT1>>=1, maskT2>>=1, pw*=0.5) {
      if (vT1 > pw) {vT1 -= pw; cd |= maskT1; zT1 = FALSE;}
      if (vT2 > pw) {vT2 -= pw; cd |= maskT2; zT2 = FALSE;}
    }
    if (zT1) cd &= ~MASK_ST1;	/* avoid "-0" */
    if (zT2) cd &= ~MASK_ST2;	/* avoid "-0" */
  }
  /* cd =	 N(sign)  N(NBPDN-2)  N(NBPDN-3) ...  N(0)	*/
  /*		T1(sign) T1(NBPDT-2) T1(NBPDT-3) ... T1(0)	*/
  /*		T2(sign) T2(NBPDT-2) T2(NBPDT-3) ... T2(0)	*/
#ifdef BIT_SHUFFLING
  mask = MASK_SN;
  if (cd & MASK_SN)	*codedelta |= mask;	mask >>= 1;
  if (cd & MASK_ST1)	*codedelta |= mask;	mask >>= 1;
  if (cd & MASK_ST2)	*codedelta |= mask;	mask >>= 1;
  for (i=0, maskN=MASK_SN>>1; i<NBPDN-NBPDT; i++, maskN>>=1) {
    if (cd & maskN)	*codedelta |= mask;	mask >>= 1;
  }
  for (i=1, maskT1=MASK_ST1>>1, maskT2=MASK_ST2>>1; i<NBPDT;
       i++, maskN>>=1, maskT1>>=1, maskT2>>=1) {
    if (cd & maskN)	*codedelta |= mask;	mask >>= 1;
    if (cd & maskT1)	*codedelta |= mask;	mask >>= 1;
    if (cd & maskT2)	*codedelta |= mask;	mask >>= 1;
  }
  /* *codedelta=N(sign) T1(sign) T2(sign) N(NBPDN-2) N(NBPDN-3)	...	*/
  /*		N(NBPDT-1) N(NBPDT-2) T1(NBPDT-2) T2(NBPDT-2)		*/
  /*		N(NBPDT-3) T1(NBPDT-3) T2(NBPDT-3) ... N(0) T1(0) T2(0)	*/
#else
  *codedelta = cd;
#endif
} /* sdmEncodeDetail */


void sdmDecodeDetail(sdMeshPtr m, int nbits, CODEWORD codedelta,
		     sdVectorPtr delta)
{
  int i;
  double vN, vT1, vT2, pw;
  CODEWORD cd, maskN, maskT1, maskT2, mask;

  delta->x = delta->y = delta->z = 0.0F;
  if (!m || !nbits)
    return;

#ifdef BIT_SHUFFLING
  /* codedelta=	N(sign) T1(sign) T2(sign) N(NBPDN-2) N(NBPDN-3)	...	*/
  /*		N(NBPDT-1) N(NBPDT-2) T1(NBPDT-2) T2(NBPDT-2)		*/
  /*		N(NBPDT-3) T1(NBPDT-3) T2(NBPDT-3) ... N(0) T1(0) T2(0)	*/
  cd = 0; mask = MASK_SN;
  if (codedelta & mask)		cd |= MASK_SN;	mask >>= 1;
  if (codedelta & mask)		cd |= MASK_ST1;	mask >>= 1;
  if (codedelta & mask)		cd |= MASK_ST2;	mask >>= 1;
  for (i=0, maskN=MASK_SN>>1; i<NBPDN-NBPDT; i++, maskN>>=1) {
    if (codedelta & mask)	cd |= maskN;	mask >>= 1;
  }
  for (i=1, maskT1=MASK_ST1>>1, maskT2=MASK_ST2>>1; i<NBPDT;
       i++, maskN>>=1, maskT1>>=1, maskT2>>=1) {
    if (codedelta & mask)	cd |= maskN;	mask >>= 1;
    if (codedelta & mask)	cd |= maskT1;	mask >>= 1;
    if (codedelta & mask)	cd |= maskT2;	mask >>= 1;
  }
#else
  cd = codedelta;
#endif
  /* cd =	 N(sign)  N(NBPDN-2)  N(NBPDN-3) ...  N(0)	*/
  /*		T1(sign) T1(NBPDT-2) T1(NBPDT-3) ... T1(0)	*/
  /*		T2(sign) T2(NBPDT-2) T2(NBPDT-3) ... T2(0)	*/
  vN = vT1 = vT2 = 0.0;
  if (nbits > NBPDN) nbits = NBPDN;
  for (i=1, maskN=MASK_SN>>1, pw=0.5*m->maxDetailN;
       i<nbits; i++, maskN>>=1, pw*=0.5)
    if (cd & maskN) vN += pw;
  if (vN > 0) vN += pw;		/* add yet another half step */
  if (cd & MASK_SN) vN = -vN;
  nbits -= NBPDN-NBPDT;
  if (nbits > 1) {
    if (nbits > NBPDT) nbits = NBPDT;
    for (i=1, maskT1=MASK_ST1>>1, maskT2=MASK_ST2>>1, pw=0.5*m->maxDetailT;
	 i<nbits; i++, maskT1>>=1, maskT2>>=1, pw*=0.5) {
      if (cd & maskT1) vT1 += pw;
      if (cd & maskT2) vT2 += pw;
    }
    if (vT1 > 0) vT1 += pw;	/* add yet another half step */
    if (cd & MASK_ST1) vT1 = -vT1;
    if (vT2 > 0) vT2 += pw;	/* add yet another half step */
    if (cd & MASK_ST2) vT2 = -vT2;
  }
  delta->x = (float) vN;  delta->y = (float) vT1;  delta->z = (float) vT2;
} /* sdmDecodeDetail */

